/*************************************************************************************************
 * xadbm.o
 * Abstraction for database managers compatible with DBM
 *                                                       Copyright(c) 2000-2003 Mikio Hirabayashi
 *************************************************************************************************/


#ifndef _XADBM_H                         /* duplication check */
#define _XADBM_H


#include <xqdbm.h>



/**
 * Error container for ADBM.
 */
class qdbm::DBM_error {
public:
  //----------------------------------------------------------------
  // public member functions
  //----------------------------------------------------------------
  /**
   * Create an instance.
   */
  DBM_error() throw();
  /**
   * Create an instance.
   * @param message the string of a error message.
   */
  DBM_error(const char* message) throw();
  /**
   * Release resources of the instance.
   */
  virtual ~DBM_error() throw();
  /**
   * Get the message.
   * @return the pointer to the string.
   */
  virtual const char* message() const throw();
private:
  //----------------------------------------------------------------
  // private member variables
  //----------------------------------------------------------------
  /** The pointer to the region of the error message. */
  char* errmsg;
};



/**
 * Datum of records for ADBM.
 */
class qdbm::Datum {
public:
  //----------------------------------------------------------------
  // public member functions
  //----------------------------------------------------------------
  /**
   * Create an instance.
   * @param dptr the pointer to the region of data.
   * @param dsize the size of the region.  If it is negative, the size is assigned
   * with `std::strlen(dptr)'.
   */
  Datum(const char* dptr = "", int dsize = -1);
  /**
   * Create an instance.
   * @param num an integer number.
   */
  Datum(int num);
  /**
   * Copy constructor.
   * @param datum a source instance.
   */
  Datum(const Datum& datum);
  /**
   * Release resources of the instance.
   */
  virtual ~Datum() throw();
  /**
   * Assignment operator.
   * @param datum a source instance.
   * @return reference to itself.
   */
  virtual Datum& operator=(const Datum& datum);
  /**
   * Assignment operator.
   * @param str a source string.
   * @return reference to itself.
   */
  virtual Datum& operator=(const char* str);
  /**
   * Concatenation operator.
   * @param datum a latter instance.
   * @return reference to itself.
   */
  virtual Datum& operator<<(const Datum& datum);
  /**
   * Concatenation operator.
   * @param str a latter string.
   * @return reference to itself.
   */
  virtual Datum& operator<<(const char* str);
  /**
   * Equality operator.
   * @param datum a comparing instance.
   * @return true if the both equal, else, false.
   */
  virtual bool operator==(const Datum& datum) const;
  /**
   * Inequality operator.
   * @param datum a comparing instance.
   * @return true if the both do not equal, else, false.
   */
  virtual bool operator!=(const Datum& datum) const;
  /**
   * Equality operator.
   * @param str a comparing string.
   * @return true if the both equal, else, false.
   */
  virtual bool operator==(const char* str) const;
  /**
   * Inequality operator.
   * @param str a comparing string.
   * @return true if the both do not equal, else, false.
   */
  virtual bool operator!=(const char* str) const;
  /**
   * Get the pointer of the region of the datum.
   * @return the pointer of the region of the datum.
   * @note Because an additional zero code is appended at the end of the region of the return
   * value, the return value can be treated as a character string.
   */
  virtual const char* ptr() const;
  /**
   * Get the size of the region of the datum.
   * @return the size of the region of the datum.
   */
  virtual int size() const;
  /**
   * Replace the contents with a C-style allocated region.
   * @param dptr the region allocated with `std::malloc' or `std::realloc'.
   * @param dsize the size of the region.
   * @note the specified region is released with `std::free' at the destruction of the instance.
   */
  virtual void cjack(char *dptr, int dsize);
private:
  //----------------------------------------------------------------
  // private member variables
  //----------------------------------------------------------------
  /** The pointer to the region. */
  char* dptr;
  /** The size of the region. */
  int dsize;
};



/**
 * Aabstraction for database managers compatible with DBM.
 */
class qdbm::ADBM {
public:
  //----------------------------------------------------------------
  // public member functions
  //----------------------------------------------------------------
  /**
   * Close the database connection.
   * @throw DBM_error if an error occures.
   */
  virtual void close() throw(DBM_error) = 0;
  /**
   * Store a record.
   * @param key reference to a key object.
   * @param val reference to a value object.
   * @param replace whether the existing value is to be overwritten or not.
   * @throw DBM_error if an error occures or replace is cancelled.
   */
  virtual void storerec(const Datum& key, const Datum& val, bool replace = true)
    throw(DBM_error) = 0;
  /**
   * Delete a record.
   * @param key reference to a key object.
   * @throw DBM_error if an error occures or no record corresponds.
   */
  virtual void deleterec(const Datum& key) throw(DBM_error) = 0;
  /**
   * Fetch a record.
   * @param key reference to a key object.
   * @return the pointer to an instance of the value of the corresponding record.
   * @throw DBM_error if an error occures or no record corresponds.
   * @note The instance pointed to by the return value is to be destloyed by the caller.
   */
  virtual Datum* fetchrec(const Datum& key) throw(DBM_error) = 0;
  /**
   * Get the first key.
   * @return the pointer to an instance of the value of the first record.
   * @throw DBM_error if an error occures or no record corresponds.
   * @note The instance pointed to by the return value is to be destloyed by the caller.
   */
  virtual Datum* firstkey() throw(DBM_error) = 0;
  /**
   * Get the next key.
   * @return the pointer to an instance of the value of the first record.
   * @throw DBM_error if an error occures or no record corresponds.
   * @note The instance pointed to by the return value is to be destloyed by the caller.
   */
  virtual Datum* nextkey() throw(DBM_error) = 0;
  /**
   * Check whether an fatal error occured or not.
   * @return true if the database has a fatal error, false if not.
   * @throw DBM_error if an error occures.
   */
  virtual bool error() throw(DBM_error) = 0;
  /**
   * Clear the error condition on the database.
   * @throw DBM_error if an error occures.
   */
  virtual void clearerror() throw(DBM_error) = 0;
};



#endif                                   /* duplication check */


/* END OF FILE */
