/*************************************************************************************************
 * xcuria.o
 * C++ API of Curia, the extended API of QDBM
 *                                                       Copyright(c) 2000-2003 Mikio Hirabayashi
 *************************************************************************************************/


#ifndef _XCURIA_H                        /* duplication check */
#define _XCURIA_H


#include <xqdbm.h>
#include <xadbm.h>

extern "C" {
#include <depot.h>
#include <curia.h>
#include <stdlib.h>
}



/**
 * Error container for Curia.
 */
class qdbm::Curia_error : public virtual DBM_error {
public:
  //----------------------------------------------------------------
  // public member functions
  //----------------------------------------------------------------
  /**
   * Create an instance.
   */
  Curia_error() throw();
  /**
   * Create an instance.
   * @param ecode the error code.
   */
  Curia_error(int ecode) throw();
  /**
   * Copy constructor.
   * @param ce a source instance.
   */
  Curia_error(const Curia_error& ce) throw();
  /**
   * Release resources of the instance.
   */
  virtual ~Curia_error() throw();
  /**
   * Assignment operator.
   * @param ce a source instance.
   * @return reference to itself.
   */
  virtual Curia_error& operator=(const Curia_error& ce) throw();
  /**
   * Assignment operator.
   * @param ecode the error code.
   * @return reference to itself.
   */
  virtual Curia_error& operator=(int ecode) throw();
  /**
   * Equality operator.
   * @param ce a comparing instance.
   * @return true if the both equal, else, false.
   */
  virtual bool operator==(const Curia_error &ce) const throw();
  /**
   * Inequality operator.
   * @param ce a comparing instance.
   * @return true if the both do not equal, else, false.
   */
  virtual bool operator!=(const Curia_error &ce) const throw();
  /**
   * Equality operator.
   * @param ecode a comparing error code.
   * @return true if the both equal, else, false.
   */
  virtual bool operator==(int ecode) const throw();
  /**
   * Inequality operator.
   * @param ecode a comparing error code.
   * @return true if the both do not equal, else, false.
   */
  virtual bool operator!=(int ecode) const throw();
  /**
   * Get the error code.
   * @return the error code.
   */
  virtual int code() const throw();
  /**
   * Get the error message.
   * @return the string.
   */
  virtual const char* message() const throw();
private:
  //----------------------------------------------------------------
  // private member variables
  //----------------------------------------------------------------
  /** The error code. */
  int ecode;
};



/**
 * C++ API of Curia, the extended API of QDBM
 */
class qdbm::Curia : public virtual ADBM {
public:
  //----------------------------------------------------------------
  // public static member constants
  //----------------------------------------------------------------
  static const int ENOERR = DP_ENOERR;         ///< error code: no error
  static const int EFATAL = DP_EFATAL;         ///< error code: with fatal error
  static const int EMODE = DP_EMODE;           ///< error code: invalid mode
  static const int EBROKEN = DP_EBROKEN;       ///< error code: broken database file
  static const int EKEEP = DP_EKEEP;           ///< error code: existing record
  static const int ENOITEM = DP_ENOITEM;       ///< error code: no item found
  static const int EALLOC = DP_EALLOC;         ///< error code: memory allocation error
  static const int EMAP = DP_EMAP;             ///< error code: memory mapping error
  static const int EOPEN = DP_EOPEN;           ///< error code: open error
  static const int ECLOSE = DP_ECLOSE;         ///< error code: close error
  static const int ETRUNC = DP_ETRUNC;         ///< error code: trunc error
  static const int ESYNC = DP_ESYNC;           ///< error code: sync error
  static const int ESTAT = DP_ESTAT;           ///< error code: stat error
  static const int ESEEK = DP_ESEEK;           ///< error code: seek error
  static const int EREAD = DP_EREAD;           ///< error code: read error
  static const int EWRITE = DP_EWRITE;         ///< error code: write error
  static const int ELOCK = DP_ELOCK;           ///< error code: lock error
  static const int EUNLINK = DP_EUNLINK;       ///< error code: unlink error
  static const int EMKDIR = DP_EMKDIR;         ///< error code: mkdir error
  static const int ERMDIR = DP_ERMDIR;         ///< error code: rmdir error
  static const int EMISC = DP_EMISC;           ///< error code: miscellaneous error
  static const int OREADER = DP_OREADER;       ///< open mode: open as a reader
  static const int OWRITER = DP_OWRITER;       ///< open mode: open as a writer
  static const int OCREAT = DP_OCREAT;         ///< open mode: writer creating
  static const int OTRUNC = DP_OTRUNC;         ///< open mode: writer truncating
  static const int DOVER = DP_DOVER;           ///< write mode: overwrite the existing value
  static const int DKEEP = DP_DKEEP;           ///< write mode: keep the existing value
  static const int DCAT = DP_DCAT;             ///< write mode: concatenate values
public:
  //----------------------------------------------------------------
  // public static member functions
  //----------------------------------------------------------------
  /**
   * Get the version information.
   * @return the string of the version information.
   */
  static const char* version() throw();
  /**
   * Remove a database directory.
   * @param name the name of a database directory.
   * @throw Curia_error if an error occures.
   */
  static void remove(const char* name) throw(Curia_error);
public:
  //----------------------------------------------------------------
  // public member functions
  //----------------------------------------------------------------
  /**
   * Get the database handle.
   * @param name the name of a database directory.
   * @param omode the connection mode: `Curia::OWRITER' as a writer, `Curia::OREADER' as
   * a reader.  If the mode is `Curia::OWRITER', the following may be added by bitwise or:
   * `Curia::OCREAT', which means it creates a new database if not exist, `Curia::OTRUNC',
   * which means it creates a new database regardless if one exists.
   * @param bnum the number of elements of each bucket array.  If it is not more than 0,
   * the default value is specified.  The size of each bucket array is determined on creating,
   * and can not be changed except for by optimization of the database.  Suggested size of each
   * bucket array is about from 0.5 to 4 times of the number of all records to store.
   * @param dnum the number of division of the database.  If it is not more than 0, the default
   * value is specified.  The number of division can not be changed from the initial value.
   * @note While connecting as a writer, an exclusive lock is invoked to the database directory.
   * While connecting as a reader, a shared lock is invoked to the database directory.
   * The thread blocks until the lock is achieved.
   */
  Curia(const char* name, int omode = Curia::OREADER, int bnum = -1, int dnum = -1)
    throw(Curia_error);
  /**
   * Release the resources.
   * @note If the database handle is not closed yet, it is closed.
   */
  virtual ~Curia() throw();
  /**
   * Close the database handle.
   * @throw Curia_error if an error occures.
   * @note Updating a database is assured to be written when the handle is closed.  If a writer
   * opens a database but does not close it appropriately, the database will be broken.
   */
  virtual void close() throw(Curia_error);
  /**
   * Store a record.
   * @param kbuf the pointer to the region of a key.
   * @param ksiz the size of the region of the key.  If it is negative, the size is assigned
   * with `std::strlen(kbuf)'.
   * @param vbuf the pointer to the region of a value.
   * @param vsiz the size of the region of the value.  If it is negative, the size is assigned
   * with `std::strlen(vbuf)'.
   * @param dmode behavior when the key overlaps, by the following values: `Curia::DOVER',
   * which means the specified value overwrites the existing one, `Curia::DKEEP', which means the
   * existing value is kept, `Curia::DCAT', which means the specified value is concatenated at
   * the end of the existing value.
   * @throw Curia_error if an error occures or replace is cancelled.
   */
  virtual void put(const char* kbuf, int ksiz, const char* vbuf, int vsiz,
                   int dmode = Curia::DOVER) throw(Curia_error);
  /**
   * Delete a record.
   * @param kbuf the pointer to the region of a key.
   * @param ksiz the size of the region of the key.  If it is negative, the size is assigned
   * with `std::strlen(kbuf)'.
   * @throw Curia_error if an error occures or no record corresponds.
   */
  virtual void out(const char* kbuf, int ksiz) throw(Curia_error);
  /**
   * Retrieve a record.
   * @param kbuf the pointer to the region of a key.
   * @param ksiz the size of the region of the key.  If it is negative, the size is assigned
   * with `std::strlen(kbuf)'.
   * @param start the offset address of the beginning of the region of the value to be read.
   * @param max the max size to read with.  If it is negative, the size is unlimited.
   * @param sp a pointer to the variable to which the size of the region of the return value
   * assigned.  If it is 0, it is not used.
   * @return the pointer to the region of the value of the corresponding record.
   * @throw Curia_error if an error occures, no record corresponds, or the size of the value
   * of the corresponding record is less than `start'.
   * @note Because an additional zero code is appended at the end of the region of the return
   * value, the return value can be treated as a character string.  Because the region of the
   * return value is allocated with the `std::malloc' call, it should be released with the
   * `std::free' call if it is no longer in use.
   */
  virtual char* get(const char* kbuf, int ksiz, int start = 0, int max = -1, int* sp = 0)
    throw(Curia_error);
  /**
   * Get the size of the value of a record.
   * @param kbuf the pointer to the region of a key.
   * @param ksiz the size of the region of the key.  If it is negative, the size is assigned
   * with `std::strlen(kbuf)'.
   * @return the size of the value of the corresponding record.
   * @throw Curia_error if an error occures or no record corresponds.
   * @note Because this function does not read the entity of a record, it is faster than `get'.
   */
  virtual int vsiz(const char* kbuf, int ksiz) throw(Curia_error);
  /**
   * Initialize the iterator of the database handle.
   * @throw Curia_error if an error occures.
   * @note The iterator is used in order to access the key of every record stored in a database.
   */
  virtual void iterinit() throw(Curia_error);
  /**
   * Get the next key of the iterator.
   * @param sp a pointer to the variable to which the size of the region of the return value
   * assigned.  If it is 0, it is not used.
   * @return the pointer to the region of the next key.
   * @throw Curia_error if an error occures or no record is to be get out of the iterator.
   * @note Because an additional zero code is appended at the end of the region of the return
   * value, the return value can be treated as a character string.  Because the region of the
   * return value is allocated with the `std::malloc' call, it should be released with the
   * `std::free' call if it is no longer in use.
   * @note It is possible to access every record by iteration of calling this function.
   * However, it is not assured if updating the database is occurred while the iteration.
   * Besides, the order of this traversal access method is arbitrary, so it is not assured that
   * the order of storing matches the one of the traversal access.
   */
  virtual char* iternext(int* sp = 0) throw(Curia_error);
  /**
   * Set alignment of the database handle.
   * @param align the basic size of alignment.
   * @throw Curia_error if an error occures.
   * @note If alignment is set to a database, the efficiency of overwriting values are improved.
   * The basic size of alignment is suggested to be average size of the values of the records
   * to be stored.  When a record is storing, the size of the region reserved for its value is
   * determined as multiple of the alignment size.  The alignment size is determined as multiple
   * of the basic size of alignment.  Because alignment setting is not saved in a database,
   * you should specify alignment every opening a database.
   */
  virtual void setalign(int align) throw(Curia_error);
  /**
   * Synchronize contents of updating the database with the files and the devices.
   * @note Curia_error if an error occures.
   * This function is useful when another process uses the connected database directory.
   */
  virtual void sync() throw(Curia_error);
  /**
   * Optimize the database.
   * @param bnum the number of the elements of each bucket array.  If it is not more than 0,
   * the default value is specified.
   * @throw Curia_error if an error occures.
   * @note In an alternating succession of deleting and storing with overwrite or concatenate,
   * dispensable regions accumulate.  This function is useful to do away with them.
   */
  virtual void optimize(int bnum = -1) throw(Curia_error);
  /**
   * Get the name of the database.
   * @return the pointer to the region of the name of the database.
   * @throw Curia_error if an error occures.
   * @note Because the region of the return value is allocated with the `std::malloc' call,
   * it should be released with the `std::free' call if it is no longer in use.
   */
  virtual char* name() throw(Curia_error);
  /**
   * Get the total size of database files.
   * @return the total size of the database files.
   * @throw Curia_error if an error occures.
   */
  virtual int fsiz() throw(Curia_error);
  /**
   * Get the total number of the elements of each bucket array.
   * @return the total number of the elements of each bucket array.
   * @throw Curia_error if an error occures.
   */
  virtual int bnum() throw(Curia_error);
  /**
   * Get the total number of the used elements of each bucket array.
   * @return the total number of the used elements of each bucket array.
   * @throw Curia_error if an error occures.
   * @note This function is inefficient because it accesses all elements of each bucket array.
   */
  virtual int busenum() throw(Curia_error);
  /**
   * Get the number of the records stored in the database.
   * @return the number of the records stored in the database.
   * @throw Curia_error if an error occures.
   */
  virtual int rnum() throw(Curia_error);
  /**
   * Check whether the database handle is a writer or not.
   * @return true if the handle is a writer, false if not.
   * @throw Curia_error if an error occures.
   */
  virtual bool writable() throw(Curia_error);
  /**
   * Check whether the database has a fatal error or not.
   * @return true if the database has a fatal error, false if not.
   * @throw Curia_error if an error occures.
   */
  virtual bool fatalerror() throw(Curia_error);
  /**
   * Get the inode number of the database directory.
   * @return the inode number of the database directory.
   * @throw Curia_error if an error occures.
   */
  virtual int inode() throw(Curia_error);
  /**
   * Store a large object.
   * @param kbuf the pointer to the region of a key.
   * @param ksiz the size of the region of the key.  If it is negative, the size is assigned
   * with `std::strlen(kbuf)'.
   * @param vbuf the pointer to the region of a value.
   * @param vsiz the size of the region of the value.  If it is negative, the size is assigned
   * with `std::strlen(vbuf)'.
   * @param dmode behavior when the key overlaps, by the following values: `Curia::DOVER',
   * which means the specified value overwrites the existing one, `Curia::DKEEP', which means the
   * existing value is kept, `Curia::DCAT', which means the specified value is concatenated at
   * the end of the existing value.
   * @throw Curia_error if an error occures or replace is cancelled.
   */
  virtual void putlob(const char* kbuf, int ksiz, const char* vbuf, int vsiz,
                      int dmode = Curia::DOVER) throw(Curia_error);
  /**
   * Delete a large object.
   * @param kbuf the pointer to the region of a key.
   * @param ksiz the size of the region of the key.  If it is negative, the size is assigned
   * with `std::strlen(kbuf)'.
   * @throw Curia_error if an error occures or no large object corresponds.
   */
  virtual void outlob(const char* kbuf, int ksiz) throw(Curia_error);
  /**
   * Retrieve a large object.
   * @param kbuf the pointer to the region of a key.
   * @param ksiz the size of the region of the key.  If it is negative, the size is assigned
   * with `std::strlen(kbuf)'.
   * @param start the offset address of the beginning of the region of the value to be read.
   * @param max the max size to be read.  If it is negative, the size to read is unlimited.
   * @param sp a pointer to the variable to which the size of the region of the return value
   * assigned.  If it is 0, it is not used.
   * @return the pointer to the region of the value of the corresponding large object.
   * @throw Curia_error if an error occures, no record corresponds, or the size of the value
   * of the corresponding record is less than `start'.
   * @note Because an additional zero code is appended at the end of the region of the return
   * value, the return value can be treated as a character string.  Because the region of the
   * return value is allocated with the `std::malloc' call, it should be released with the
   * `std::free' call if it is no longer in use.
   */
  virtual char* getlob(const char* kbuf, int ksiz, int start = 0, int max = -1, int* sp = 0)
    throw(Curia_error);
  /**
   * Get the size of the value of a large object.
   * @param kbuf the pointer to the region of a key.
   * @param ksiz the size of the region of the key.  If it is negative, the size is assigned
   * with `std::strlen(kbuf)'.
   * @return the size of the value of the corresponding large object.
   * @throw Curia_error if an error occures or no large object corresponds.
   * @note Because this function does not read the entity of a large object, it is faster than
   * `getlob'.
   */
  virtual int vsizlob(const char* kbuf, int ksiz) throw(Curia_error);
  /**
   * Get the number of the large objects stored in the database.
   * @return the number of the large objects stored in the database.
   * @throw Curia_error if an error occures.
   */
  virtual int rnumlob() throw(Curia_error);
  /**
   * Store a record.
   * @param key reference to a key object.
   * @param val reference to a value object.
   * @param replace whether the existing value is to be overwritten or not.
   * @throw Curia_error if an error occures or replace is cancelled.
   */
  virtual void storerec(const Datum& key, const Datum& val, bool replace = true)
    throw(Curia_error);
  /**
   * Delete a record.
   * @param key reference to a key object.
   * @throw Curia_error if an error occures or no record corresponds.
   */
  virtual void deleterec(const Datum& key) throw(Curia_error);
  /**
   * Fetch a record.
   * @param key reference to a key object.
   * @return the pointer to an instance of the value of the corresponding record.
   * @throw Curia_error if an error occures or no record corresponds.
   * @note The instance pointed to by the return value is to be destloyed by the caller.
   */
  virtual Datum* fetchrec(const Datum& key) throw(Curia_error);
  /**
   * Get the first key.
   * @return the pointer to an instance of the value of the first record.
   * @throw Curia_error if an error occures or no record corresponds.
   * @note The instance pointed to by the return value is to be destloyed by the caller.
   */
  virtual Datum* firstkey() throw(Curia_error);
  /**
   * Get the next key.
   * @return the pointer to an instance of the value of the first record.
   * @throw Curia_error if an error occures or no record corresponds.
   * @note The instance pointed to by the return value is to be destloyed by the caller.
   */
  virtual Datum* nextkey() throw(Curia_error);
  /**
   * Check whether an fatal error occured or not.
   * @return true if the database has a fatal error, false if not.
   * @throw Curia_error if an error occures.
   */
  virtual bool error() throw(Curia_error);
  /**
   * No effect.
   * @throw Curia_error if an error occures.
   */
  virtual void clearerror() throw(Curia_error);
private:
  //----------------------------------------------------------------
  // private member variables
  //----------------------------------------------------------------
  /** The database handle. */
  CURIA *curia;
  //----------------------------------------------------------------
  // private member functions
  //----------------------------------------------------------------
  /** copy constructor: This should not be used. */
  Curia(const Curia& curia) throw(Curia_error);
  /** assignment operator: This should not be used. */
  virtual Curia& operator=(const Curia& curia) throw(Curia_error);
};



#endif                                   /* duplication check */


/* END OF FILE */
